﻿<#
指定されたストア内の vhdx ファイルを確認して、親 vhdx のファイル名が
正しくなかった場合には修正します。

パラメーター

  -store        ストアパスの場所。
  -dryrun       修正する必要があるか確認しますが、実際には修正しません。
  -v            詳細なログ表示を有効にします。

#>

[CmdletBinding()]
param(
    [Parameter(Mandatory=$true,HelpMessage="ストアパス")]
    [string]$store,

    [switch]$dryrun
)


# Write-Error で止まるようにする
$ErrorActionPreference = "Stop"

function Main() {
    if (!(([Security.Principal.WindowsPrincipal] [Security.Principal.WindowsIdentity]::GetCurrent()).IsInRole(`
            [Security.Principal.WindowsBuiltInRole] "Administrator"))) {
        Write-Error "管理者として実行してください。"
    }

    if (-not (Test-Path -PathType container -Path $store)) {
        Write-Error "ストアパス $store が存在しません。"
    }

    Write-Host "ストア: $store"
    $files = @(Get-FilesToFix $store)
    Write-Host

    if ($files.Count -eq 0) {
        Write-Host "修正する必要はありません。"
        return
    }

    if ($dryRun) {
        Write-Host "$($files.Count) 個のファイルを修正する必要があります。"
        Write-Host "dryrun モードのため、ここで終了します。"
        return
    }

    Write-Host "$($files.Count) 個のファイルを修正します..."
    foreach ($file in $files) {
        Fix-VhdxParent $file.Path $file.ParentGuid
    }
    Write-Host "修正が完了しました。	"
}


function Get-FilesToFix($store) {
    <#
    ストアパスから Parent Locator を修正すべき vhdx ファイルを取得します。

    戻り値の例
    @(
        [PSCustomObject]@{Path = "C:\xxxx.vhdx"; ParentGuid = "GUID"},
    )
    #>
    $pathes = @(Get-ChildItem (Join-Path $store "*__*__*.vhdx") | `
        Where-Object { $_.Name -imatch "^[0-9a-f]{8}-(?:[\da-f]{4}-){3}[\da-f]{12}__[\da-f]{8}-(?:[0-9a-f]{4}-){3}[\da-f]{12}__\d+\.vhdx$" })
    Write-Host "$($pathes.Count) 個の vhdx を確認します..."

    $ret = @()
    foreach ($path in $pathes) {
        # ファイル名から情報を取得する
        $parts = $path.Name -split "__"
        if ($parts[2] -eq "0.vhdx") {
            continue
        }
        $expectedParent = $parts[1]

        # vhdx の情報を取得する
        Write-Verbose "    Checking $($path.Name)..."
        $vhdx = Get-VHD $path.FullName

        # 親がないファイルは確認対象外
        if ($vhdx.ParentPath -eq "") {
            Write-Verbose "    -> No parent"
            continue
        }

        # 親がいるとき
        $parentFileName = Split-Path $vhdx.ParentPath -Leaf
        if ($parentFileName -eq "$expectedParent.vhdx") {
            Write-Verbose "    -> No problem"
        }
        else {
            Write-Host "    $($path.Name) : 親 が $parentFileName"
            $ret += [PSCustomObject]@{Path = $path; ParentGuid = $expectedParent}
        }
    }
    Write-Host "確認完了しました。"
    return $ret
}


function Fix-VhdxParent($path, $guid) {
    <#
    指定されたファイルの親情報を修正します。
    #>
    # ファイルの情報を取得する
    $fileObj = Get-Item $path
    $modifiedTime = $fileObj.LastWriteTime
    $dir = Split-Path -Path $path
    $filename = Split-Path -Path $file.Path -Leaf

    Write-Verbose "    $filename の親を $guid.vhdx に変更します..."
    Set-ItemProperty $file.Path -Name IsReadOnly -Value $false
    try {
        # vhd_tool で親を書き換える
        $process = New-Object System.Diagnostics.Process
        $process.StartInfo.FileName = "$PSScriptRoot\vhd_tool.exe"
        $process.StartInfo.Arguments = "--redirect ""$path"" .\$guid.vhdx ""$dir\$guid.vhdx"" $guid"
        $process.StartInfo.UseShellExecute = $false
        $process.StartInfo.CreateNoWindow = $true
        $process.StartInfo.UseShellExecute = $false
        $process.StartInfo.RedirectStandardOutput = $true
        $process.StartInfo.RedirectStandardError = $true

        $out = New-Object System.Text.StringBuilder
        $err = New-Object System.Text.StringBuilder
        $callback = {
           if (![String]::IsNullOrEmpty($EventArgs.Data)) {
               $Event.MessageData.AppendLine($EventArgs.Data)
           }
        }
        $evt1 = Register-ObjectEvent -InputObject $process -Action $callback -EventName "OutputDataReceived" -MessageData $out
        $evt2 = Register-ObjectEvent -InputObject $process -Action $callback -EventName "ErrorDataReceived" -MessageData $err
        try {
            if (-not $process.Start()) {
                Write-Warning "    vhd_tool.exe を実行できませんでした。$PSScriptRoot に配置してあるか確認してください。"
                exit 1
            }
            $process.BeginOutputReadLine()
            $process.BeginErrorReadLine()
            $process.WaitForExit()
        }
        finally {
            Unregister-Event -SourceIdentifier $evt1.Name
            Unregister-Event -SourceIdentifier $evt2.Name
        }
        if ($process.ExitCode -ne 0) {
            Write-Warning "    $filename の修正に失敗しました"
            Write-Host "stdout: $($out.ToString())"
            Write-Host "stderr: $($err.ToString())"
            exit 1
        }
        Write-Host "    $filename を修正しました。"
        Write-Verbose "stdout: $($out.ToString())"
        Write-Verbose "stderr: $($err.ToString())"
    }
    finally {
        # 更新日時を戻して、read only にする
        $fileObj.LastWriteTime = $modifiedTime
        Set-ItemProperty $file.Path -Name IsReadOnly -Value $true
    }
}


try {
    Write-Host
    Main
    Write-Host
}
catch {
    Write-Host $_.Exception.Message -ForegroundColor Red
    Write-Verbose $_.ScriptStackTrace 
    Write-Host
    exit 1
}
